/*
* THIS SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS.  BITFLASH GRAPHICS INC.
* (BITFLASH) ASSUMES NO RESPONSIBILITY FOR THE ACCURACY OR COMPLETENESS OF THE
* DOCUMENTATION AND EXPRESSLY DISCLAIMS ALL REPRESENTATIONS AND WARRANTIES OF ANY
* KIND REGARDING THE CONTENT OR USE OF THE SOFTWARE OR DOCUMENTATION INCLUDING,
* BUT NOT LIMITED TO, EXPRESS AND IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
* FOR A PARTICULAR USE, OR NON-INFRINGEMENT.
*
* IN NO EVENT WILL BITFLASH BE LIABLE FOR ANY DIRECT, INDIRECT, SPECIAL,
* INCIDENTAL OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, LOST BUSINESS OR
* LOST DATA, RESULTING FROM THE USE OF OR RELIANCE UPON THE SOFTWARE OR
* DOCUMENTATION, WHETHER OR NOT BITFLASH HAS BEEN ADVISED OF THE POSSIBILITY OF
* SUCH DAMAGES.
*/

package com.bitflash.jabber.client.connection;

import java.util.*;

import com.bitflash.jabber.client.model.*;
import com.bitflash.jabber.client.model.jabber.*;

/** 
 * Represents a Jabber connection to the rest of the application.
 */
public class JabberConnection implements JabberStreamListener
{	
	public JabberConnection( 
		JabberListener jabberListener,
		String userName,
		String password,
		String server,
		String resource
	)
	{
		this.jabberListener = jabberListener;
		
		this.userName = userName;
		this.password = password;
		this.server = server;
		this.resource = resource;
		
		this.sendQueue = new Vector();
		
		this.stream = new JabberStream( this, this.server );
	}
	
	public boolean open() 
		{ return stream.open(); }
	
	public boolean close()
		{ return stream.close(); }
	
	public String jabberID()
		{ return Authentication.generateJabberID( userName, server, resource ); }
	
	public void send( DocumentObject element )
	{		
		sendQueue.addElement( element );
		stateMachine( eventSEND, element );
	}
		
	// ------------------------------------- JabberStreamListener
	
	public void handleJabberEvent( DocumentObject element )
	{
		// Do a little preprocessing before we go to the state
		// machine.
		
		if( element instanceof com.bitflash.jabber.client.model.jabber.Error )
			stateMachine( eventERROR, element );
		
		else if( element instanceof InfoQuery )
		{
			if( element.getProperty( "type").equals( "result" )	)
				stateMachine( eventRESULT, element );
		}
		
		else if( element instanceof Roster )	
			stateMachine( eventROSTER, element );
		
		else if( element instanceof Presence )
			stateMachine( eventPRESENCE, element );
		
		else if( element instanceof Message )
			stateMachine( eventMESSAGE, element );		
	}
	
	public void jabberStreamOpened() 
	{
		stateMachine( eventSTREAM_OPENED, null );
	}
	
	public void jabberStreamClosed() 
	{
		stateMachine( eventSTREAM_CLOSED, null );
	}
		
	// ------------------------------------- Finite State Machine			
	
	private synchronized void stateMachine( int type, Object data )
	{	
		int oldState = state;
		
		try
		{
			switch( state )
			{
				case START:
					stateSTART( type, data );
					break;
					
				case AUTHENTICATING:
					stateAUTHENTICATING( type, data );
					break;
					
				case NORMAL:
					stateNORMAL( type, data );
					break;
					
				case STOPPED:
					break;
				
				default:
					throw new Exception( "Unknown state" );					
			}
		} 
		catch( Throwable t )
		{
			t.printStackTrace();
			state = STOPPED;
		}
		
		if( oldState != state && STOPPED == state )
			jabberListener.jabberStopped();
	}
					
	private void stateSTART( int type, Object data ) throws Exception
	{
		switch( type )
		{
			case eventSTREAM_OPENED:
			{
				stream.send( 
					Authentication.create( userName, password, resource ) 
				);		
				state = AUTHENTICATING;
				break;
			}								
			
			case eventSTREAM_CLOSED:
			{
				state = STOPPED;
				break;
			}
		}				
	}
		
	private void stateAUTHENTICATING( int type, Object data ) throws Exception
	{
		switch( type )		
		{
			case eventRESULT:
			{
				stream.send( Presence.create( userName, server, resource ) );
				state = NORMAL;
				jabberListener.jabberStarted();
				
				sendAllQueued();
				break;
			}
			
			case eventERROR:
			{
				stream.close();
				break;
			}
						
			case eventSTREAM_CLOSED:
			{
				state = STOPPED;
				break;
			}
		}						
	}
					
	private void stateNORMAL( int type, Object data ) throws Exception
	{		
		switch( type )
		{
			case eventSEND:
			{
				sendAllQueued();
				break;				
			}
			
			case eventPRESENCE:
			{
				jabberListener.jabberPresence( (Presence)data );
				break;
			}
			
			case eventROSTER:
			{
				jabberListener.jabberRoster( (Roster)data );
				break;
			}
			
			case eventMESSAGE:
			{
				jabberListener.jabberMessage( (Message)data );
				break;
			}						
			
			case eventSTREAM_CLOSED:
			{
				state = STOPPED;
				break;
			}
		}
	}
	
	private void sendAllQueued() throws Exception
	{
		Enumeration E = sendQueue.elements();
		while( E.hasMoreElements() )
			stream.send( (DocumentObject)E.nextElement() );
		sendQueue = new Vector();		
	}
	
	JabberStream stream;
	JabberListener jabberListener;
		
	String userName;
	String password;	
	String server;
	String resource;
	
	Vector sendQueue;
	
	static final int START = 0;
	static final int STOPPED = 1;
	static final int AUTHENTICATING = 2;
	static final int NORMAL = 3;
	
	static final int eventERROR = 0;
	static final int eventSTREAM_OPENED = 1;
	static final int eventSTREAM_CLOSED = 2;
	static final int eventSEND = 3;
	static final int eventRESULT = 4;	
	static final int eventPRESENCE = 5;
	static final int eventROSTER = 6;
	static final int eventMESSAGE = 7;	
	
	int state = START;
}
