/*
* THIS SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS.  BITFLASH GRAPHICS INC.
* (BITFLASH) ASSUMES NO RESPONSIBILITY FOR THE ACCURACY OR COMPLETENESS OF THE
* DOCUMENTATION AND EXPRESSLY DISCLAIMS ALL REPRESENTATIONS AND WARRANTIES OF ANY
* KIND REGARDING THE CONTENT OR USE OF THE SOFTWARE OR DOCUMENTATION INCLUDING,
* BUT NOT LIMITED TO, EXPRESS AND IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
* FOR A PARTICULAR USE, OR NON-INFRINGEMENT.
*
* IN NO EVENT WILL BITFLASH BE LIABLE FOR ANY DIRECT, INDIRECT, SPECIAL,
* INCIDENTAL OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, LOST BUSINESS OR
* LOST DATA, RESULTING FROM THE USE OF OR RELIANCE UPON THE SOFTWARE OR
* DOCUMENTATION, WHETHER OR NOT BITFLASH HAS BEEN ADVISED OF THE POSSIBILITY OF
* SUCH DAMAGES.
*/

package com.bitflash.jabber.client.connection;

import org.xml.sax.*;
import java.io.*;
import java.net.*;
import java.util.*;

import com.bitflash.jabber.client.model.*;
import com.bitflash.jabber.client.model.jabber.*;
import com.bitflash.jabber.client.xml.*;

/**
 * Talk to the Jabber server.
 */
class JabberStream extends HandlerBase implements Runnable
{
	static final int DEFAULT_PORT = 5222;
	static final boolean TRACE = true;

	String server;
	int port;
		
	Socket theSocket;
	DataInputStream input;
	DataOutputStream output;
	
	JabberStreamListener eventListener;
	Stack elementStack;
		
	public JabberStream( JabberStreamListener eventListener, String server ) 
	{
		this( eventListener, server, DEFAULT_PORT );
	}
	
	public JabberStream( JabberStreamListener eventListener, String server, int port )
	{		
		this.eventListener = eventListener;
		this.server = server;
		this.port = port;
		this.elementStack = new Stack();
	}	
	
	public boolean open()
	{
		try
		{
			theSocket = new Socket( server, port );
					
			input = new DataInputStream( theSocket.getInputStream() );
			output = new DataOutputStream( theSocket.getOutputStream() );

			String hello = 				
				"<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n\n" +
				"<stream:stream to=\"" + server + "\" xmlns=\"jabber:client\" " + 
				"xmlns:stream=\"http://etherx.jabber.org/streams\">\n\n";
			
			if( TRACE ) System.out.println( "OUT: " + hello );
			
			output.writeBytes( hello );
			
			new Thread(this).start();
		}
		catch( Throwable t )
		{
			t.printStackTrace();
			
			theSocket = null;
			input = null;
			output = null;
			
			return false;
		}
		return true;
	}
	
	synchronized public boolean close()
	{
		if( null == theSocket )
			return true;
		
		try
		{
			output.writeBytes( "</stream:stream>" );
		} 
		catch( Throwable t )
		{
			t.printStackTrace();
			return false;
		}
		return true;
	}
	
	public void send( DocumentObject element ) throws IOException
	{		
		if( TRACE ) { System.out.println( "OUT: " + element.toString() + "\n" ); }
		output.writeBytes( element.toString() );
	}
	
	// ----------------------------------------- XML (SAX) Parser
	
    public void startElement (String name, AttributeList attributes)
		throws SAXException
    {
		if( TRACE )
		{
			INDENT(); System.out.println( "<" + name );
			for( int i = 0; i < attributes.getLength(); i++ )
			{
				INDENT(); INDENT(1); System.out.println( attributes.getName(i) + "=\"" + attributes.getValue(i) + "\"" );
			}
			INDENT(); System.out.println( ">");
		}
		
		if( name.equals( "stream:stream" ) )
			eventListener.jabberStreamOpened();
		
		else
		{
			XMLElement element = new XMLElement( name );
			for( int i = 0; i < attributes.getLength(); i++ )
				element.setAttribute( attributes.getName(i), StreamingXMLParser.decodeUTF(attributes.getValue(i)) );						
			elementStack.push( element );
		}
    }

    public void endElement (String name)
		throws SAXException
    {		
		// Connection closed. </stream:stream> 
		if( elementStack.isEmpty() )
			return;
		
		XMLElement xmlElement = (XMLElement)elementStack.pop();
		
		if( elementStack.isEmpty() ) 
		{
			DocumentObject compiled = compile( xmlElement );
			if( null != compiled )
				eventListener.handleJabberEvent( compiled );
		} 
		else 
		{
			((XMLElement)elementStack.peek()).addChild( xmlElement );
		}
		
		if( TRACE ) { INDENT(); System.out.println( "</" + name + ">" ); }
    }

    public void characters (char buf [], int offset, int len)
		throws SAXException
    {	
		if( TRACE ) { INDENT(); System.out.println( "CDATA: |" +  new String( buf, offset, len ) + "|" ); }
		
		if( !elementStack.isEmpty() )
		{
			((XMLElement)elementStack.peek()).appendCharacterData( 
				StreamingXMLParser.decodeUTF(new String( buf, offset, len ))
			);
		}
    }
	
	public void run()
	{
		try {			
			// Use the default (non-validating) parser
			StreamingXMLParser parser = new StreamingXMLParser( input, this );
			parser.start();		
		} catch (Throwable t) {
		    t.printStackTrace ();
		}
		
		synchronized( this )
		{
			try	{
				output.close();
				input.close();
			} catch( Throwable t ) {
				t.printStackTrace ();
			}
		
			try {		
				theSocket.close();		
			} catch( Throwable t ) {
				t.printStackTrace();
			}
							
			theSocket = null;
			output = null;
			input = null;
		}
		
		eventListener.jabberStreamClosed();
	}
	
	// ------------------------------------- XML Element Compiler
	
	private DocumentObject compile( XMLElement element )
	{
		if( element.getName().equals( "iq" ) )
			return InfoQuery.compile( element );
		
		else if( element.name.equals( "presence" ) )
			return Presence.compile( element );		
		
		else if( element.name.equals( "message" ) )
			return Message.compile( element );
					
		return null;
	}	
	
	private void INDENT() {	if(TRACE) INDENT(elementStack.size()); }
	private void INDENT( int depth ) {	if(TRACE) while( depth-- != 0 ) System.out.print( "    " ); }
}
