/*
* THIS SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS.  BITFLASH GRAPHICS INC.
* (BITFLASH) ASSUMES NO RESPONSIBILITY FOR THE ACCURACY OR COMPLETENESS OF THE
* DOCUMENTATION AND EXPRESSLY DISCLAIMS ALL REPRESENTATIONS AND WARRANTIES OF ANY
* KIND REGARDING THE CONTENT OR USE OF THE SOFTWARE OR DOCUMENTATION INCLUDING,
* BUT NOT LIMITED TO, EXPRESS AND IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
* FOR A PARTICULAR USE, OR NON-INFRINGEMENT.
*
* IN NO EVENT WILL BITFLASH BE LIABLE FOR ANY DIRECT, INDIRECT, SPECIAL,
* INCIDENTAL OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, LOST BUSINESS OR
* LOST DATA, RESULTING FROM THE USE OF OR RELIANCE UPON THE SOFTWARE OR
* DOCUMENTATION, WHETHER OR NOT BITFLASH HAS BEEN ADVISED OF THE POSSIBILITY OF
* SUCH DAMAGES.
*/

package com.bitflash.jabber.client.gui;
import com.bitflash.jabber.client.model.jabber.*;
import java.util.*;

public abstract class ChatManager
{	
	/**
	 * @param type either CHAT or GROUPCHAT
	 */
	ChatManager( Main parent )
	{
		this.parent = parent;
		this.windows = new Vector();
	}
	
	/**
	 * Create a new window/chat session to the given 
	 * Jabber ID (could be a group ID).
	 */
	abstract ChatWindow createSession( String remoteJabberID );
	
	protected ChatWindow createSession( String remoteJabberID, String threadID )
	{
		synchronized( windows )
		{
			ChatWindow window = new ChatWindow( this, remoteJabberID, threadID );
			windows.addElement( window );
			return window;
		}		
	}

	/**
	 * Both groupchat and whiteboard presences are relevant to
	 * us. However, they may not be relevant to other people.
	 * Return true if we handle it.
	 */
	abstract boolean jabberPresence( Presence presence );
	
	void jabberMessage( Message message )
	{
		String from = message.fromNickname();
		String threadID = (String)message.getProperty("thread");
		
		ChatWindow target = null;
		
		synchronized( windows )
		{		
			Enumeration E = windows.elements();
			while( E.hasMoreElements() )
			{
				ChatWindow window = (ChatWindow)E.nextElement();
				if( window.to.equals(from) 
					&& window.threadID.equals(threadID) )
				{
					target = window;
					break;
				}
			}
		}
		
		if( null == target )
			unknownSession( message );
		
		else
			dispatchMessage( target, message );
	}
	
	/**
	 * Handle a message that's being sent to an unknown session.
	 */
	abstract void unknownSession( Message message );
	
	void dispatchMessage( ChatWindow window, Message message )
	{
		String from = (String)message.getProperty("from");
		String body = (String)message.getProperty("body");
		
		if( !body.equals("") )
			window.jabberMessage( nickname(from), body );
		
		if( null != message.getChild("x", X.XMLNS_WHITEBOARD) )
			window.whiteboardCommand( message );
	}
	
	public void sendMessage( ChatWindow window, String body )
	{
		if( body.equals("") )
			return;

		Message message = createMessage( window );
		message.setProperty( "body", body );
		parent.send( message );
	}
	
	public void sendWhiteboardCommand( ChatWindow window, X command )
	{
		Message message = createMessage( window );
		message.addChild( command );		
		parent.send( message );		
	}
		
	protected Message createMessage( ChatWindow window )
	{
		Message message = new Message();
		
		message.setProperty( "from", jabberID() );
		message.setProperty( "to", window.to );
		message.setProperty( "type", getType() );
		message.setProperty( "thread", window.threadID );

		return message;
	}
		
	public void windowClosed( ChatWindow window )
	{
		windows.removeElement( window );
	}
	
	public void closeAll()
	{
		synchronized( windows )
		{
			Enumeration E = windows.elements();
			while( E.hasMoreElements() )
			{
				ChatWindow window = (ChatWindow)E.nextElement();
				window.dispose();
			}
		}
	}
	
	/**
	 * The full, uncensored Jabber ID. See also whoAmI().
	 */
	public String jabberID()
	{
		return parent.jabberID();
	}
		
	abstract protected String getType();

	// Returns the prefix of the jabberID. The bit before the '@'.
	public String bareNickname( String jabberID )
	{
		int at = jabberID.lastIndexOf('@');
		if( at < 0 )
			return jabberID;
		return jabberID.substring( 0, jabberID.lastIndexOf('@') );
	}
	
	/**
	 * Extracts the nickname (handle) from the given Jabber ID
	 */
	abstract public String nickname( String jabberID );
	
	Main parent;
	Vector windows;
}
