/*
* THIS SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS.  BITFLASH GRAPHICS INC.
* (BITFLASH) ASSUMES NO RESPONSIBILITY FOR THE ACCURACY OR COMPLETENESS OF THE
* DOCUMENTATION AND EXPRESSLY DISCLAIMS ALL REPRESENTATIONS AND WARRANTIES OF ANY
* KIND REGARDING THE CONTENT OR USE OF THE SOFTWARE OR DOCUMENTATION INCLUDING,
* BUT NOT LIMITED TO, EXPRESS AND IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
* FOR A PARTICULAR USE, OR NON-INFRINGEMENT.
*
* IN NO EVENT WILL BITFLASH BE LIABLE FOR ANY DIRECT, INDIRECT, SPECIAL,
* INCIDENTAL OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, LOST BUSINESS OR
* LOST DATA, RESULTING FROM THE USE OF OR RELIANCE UPON THE SOFTWARE OR
* DOCUMENTATION, WHETHER OR NOT BITFLASH HAS BEEN ADVISED OF THE POSSIBILITY OF
* SUCH DAMAGES.
*/

package com.bitflash.jabber.client.gui;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import com.bitflash.jabber.client.connection.*;
import com.bitflash.jabber.client.model.jabber.*;
import com.bitflash.jabber.client.xml.*;

public class Main 
	extends Frame 
	implements 
		ActionListener, 
		ItemListener, 
		WindowListener, 
		JabberListener
{		
	public Main()	
	{		
		contactList = new Vector();
		eventList = new Vector();
		privateChatManager = new PrivateChatManager( this );
		groupChatManager = new GroupChatManager( this );
		initializeGUI();
	}			
				
	public void jabberStarted() 
	{
		// Get the roster.
		send( Roster.createRequest() );
	}
	
	public void jabberStopped() 
	{
		
		disconnect(); 
	}
	
	public void jabberPresence( Presence presence ) 
	{
		if( privateChatManager.jabberPresence(presence) )
			return;
		
		if( groupChatManager.jabberPresence(presence) )
			return;
		
			// Huh? Must be something strange. (like a whiteboard presence)
			String type = (String)presence.getProperty( "type" );
			if( type.equals("") )
				return;
		
			// Update our internal representation of the contact list
			String from = presence.fromNickname();		
		
		synchronized( contactList )
		{		
			// Try to find the person amongst the existing ones.
			int contact = findContact( from );
			if( contact >= 0 )
			{
				contactList.setElementAt( presence, contact );			
				updateContactList();
			}
		}
	}
	
	public void jabberRoster( Roster roster )
	{	
		synchronized( contactList )
		{		
			Enumeration E = roster.children.elements();
			while( E.hasMoreElements() )
			{
				XMLElement child = (XMLElement)E.nextElement();
				if( child instanceof Roster.Item )
				{
					Presence presence = ((Roster.Item)child).asPresence();
					if( findContact(presence.fromNickname()) < 0 )
						contactList.addElement( presence );
				}
			}
		
			updateContactList();
		}
	}		
	
	public void jabberMessage( Message message ) 
	{
		String type = (String)message.getProperty("type");
		if( type.equals("chat") )
		{
			privateChatManager.jabberMessage( message );
			return;
		}
		
		if( type.equals("groupchat") )
		{
			groupChatManager.jabberMessage( message );
			return;
		}
		
		eventList.insertElementAt( message, 0 );
		eventTable.addRow( new String[] { message.fromNickname(), (String)message.getProperty("subject") }, 0 );
		eventText.setText( (String)message.getProperty("body") );
	}			
		
	public void actionPerformed( ActionEvent event )
	{
		if( event.getSource() instanceof MenuItem )
		{		
			String source = ((MenuItem)event.getSource()).getLabel();
			if( "Jabber" == source )
				menuJabber( event.getActionCommand() );
		}
		
		if( event.getSource() instanceof Component )
		{
			if( null == connection )
				return;
			
			// Send message to someone on the contact list.
			String source = ((Component)event.getSource()).getName();
			if( "contactList" == source )
			{
				MessageDialog dialog = null;
				
				synchronized( contactList )
				{				
					dialog = new MessageDialog( 
						this, 
						connection.jabberID(), 
						contactListTable.getSelectedItem()[0]
					);
				}
				if( dialog.open() )
					send( dialog.createMessage() );
			}
			
			// Reply to a message on the event list.
			if( "eventTable" == source  )
			{
				Message replyTo = (Message)eventList.elementAt( eventTable.getSelectedIndex() );
				
				MessageDialog dialog = new MessageDialog(
					this,
					connection.jabberID(),
					(String)replyTo.getProperty("from")
				);
				dialog.subject = (String)replyTo.getProperty("subject");
				
				if( dialog.open() )
					send( dialog.createMessage() );
			}
		} 
	}
	
	public void itemStateChanged( ItemEvent event )
	{	
		String source = ((Component)event.getSource()).getName();
		
		if( "eventTable" == source )
		{
			int selected = eventTable.getSelectedIndex();
			eventText.setText( (String)((Message)eventList.elementAt(selected)).getProperty("body") );
		}
	}
	
	private void menuJabber( String command )
	{
		if( "Connect/Disconnect..." == command )
		{				
			if( connection != null )
			{							
				disconnect();
				return;
			}						
			
			ConnectionDialog dialog = new ConnectionDialog(this);
			if( dialog.open() )
			{
				try				
				{	
					synchronized( contactList )
					{
						// Clear out contacts. We'll reacquire this information
						// from the Roster and Presence information.
						contactList = new Vector();
						updateContactList();
					}
					
					connection = new JabberConnection( 
						this, 
						dialog.userName, 
						dialog.password, 
						dialog.server, 
						dialog.resource 
					);					
					if( !connection.open() )
					{
						connection = null;
						return;
					}
					
					setTitle( connection.jabberID() );
				}
				catch( Throwable t )
				{
					t.printStackTrace();
					connection = null;
				}
			}
		}
		
		else if( "Chat" == command && null != connection )
		{
			synchronized( contactList )
			{
				int selected = contactListTable.getSelectedIndex();
				if( selected >= 0 )
					privateChatManager.createSession( ((Presence)contactList.elementAt(selected)).fromNickname() );
			}
		}
		
		else if( "Group Chat" == command && null != connection )
		{
			GroupChatDialog dialog = new GroupChatDialog(this);
			if( dialog.open() )
			{
				groupChatManager.createSession( dialog.group + '@' + dialog.server );
			}
		}		
	}	

	public void send( com.bitflash.jabber.client.model.DocumentObject object )
	{
		try
		{
			connection.send( object );
		}
		catch( NullPointerException e )
		{
			// Not connected. This is the most threadsafe way to deal with 
			// this case.
		}
	}
	
	public String jabberID()
	{
		// Use try/catch because it's the most thread safe way
		// to do this (better than if( null == connection ) ...)
		try
		{
			return connection.jabberID();
		}
		catch( NullPointerException e )
		{
			// Not connected.
			return "";
		}		
	}
	
	private void disconnect()
	{				
		if( null == connection )
			return;
		
		privateChatManager.closeAll();
		groupChatManager.closeAll();
		
		try {
			connection.close();
		} catch( Throwable t ) {
			t.printStackTrace();
		}
		connection = null;

		synchronized( contactList )
		{		
			// Make the contact list reflect our lack of knowledge.		
			for( int i = 0; i < contactList.size(); i++ )
			{
				((Presence)contactList.elementAt(i)).setProperty( "show", "" );
				((Presence)contactList.elementAt(i)).setProperty( "status", "unknown" );
				((Presence)contactList.elementAt(i)).setProperty( "type", "unavailable" );
			}		
			updateContactList();
		}
		
		setTitle("");
	}
			
	private void initializeGUI()
	{
		setResizable(true);
		setLayout( new GridLayout(1,2) );
		addWindowListener(this);
		setBackground( Color.white );
		setForeground( Color.black );

		contactListTable = new Table();
		contactListTable.setName( "contactList" );
		
		List column = contactListTable.addColumn();
		column.setBackground( Color.blue );
		column.setForeground( Color.white );
		
		column = contactListTable.addColumn();
		column.setBackground( Color.blue );
		column.setForeground( Color.white );				
		
		contactListTable.addActionListener( this );			
		add( contactListTable );
			
		Panel messagePanel = new Panel();
		messagePanel.setLayout( new GridLayout(2,1) );
		
			eventTable = new Table();
			eventTable.addColumn();
			eventTable.addColumn();
			eventTable.addActionListener( this );
			eventTable.addItemListener( this );
			eventTable.setName( "eventTable" );								
			messagePanel.add( eventTable );
		
			eventText = new TextArea();
			messagePanel.add( eventText );
				
		add( messagePanel );
		
		MenuBar menuBar = new MenuBar();
		Menu menu = new Menu();
		menu.setLabel( "Jabber" );			
		menu.add( "Connect/Disconnect..." );
		menu.add( "Chat" );
		menu.add( "Group Chat" );
		menu.addActionListener(this);
		menuBar.add( menu );		
		setMenuBar( menuBar );
		
		pack();
		show();
	}
	
	private int findContact( String jabberID )
	{
		synchronized( contactList )
		{
			String nickname = Presence.extractNickname(jabberID);
		
			for( int i = 0; i < contactList.size(); i++ )
				if( ((Presence)contactList.elementAt(i)).fromNickname().equals( nickname ) )
					return i;
		
			return -1;
		}
	}	
	
	private void updateContactList()
	{
		synchronized( contactList )
		{
			int selected = contactListTable.getSelectedIndex();		
			contactListTable.removeAll();
		
			Enumeration E = contactList.elements();
			while( E.hasMoreElements() )
			{
				Presence presence = (Presence)E.nextElement();
				contactListTable.addRow( new String[] { 
					presence.fromNickname(),
					presence.status()
				});
			}
		
			contactListTable.select( selected );
		}
	}		
		
	public synchronized void windowClosed(WindowEvent event)
		{ disconnect(); System.exit(0); }

	public synchronized void windowClosing(WindowEvent event)
		{ disconnect(); System.exit(0); }
	
	public void windowDeiconified(WindowEvent event){}
	public void windowIconified(WindowEvent event){}
	public void windowActivated(WindowEvent event){}
	public void windowDeactivated(WindowEvent event){}
	public void windowOpened(WindowEvent event){}		

	public static void main(String[] args) 
	{		
		try { 
			new Main(); 
		} catch( Throwable t ) { 
			t.printStackTrace(); 
		}
	}	

	ChatManager privateChatManager;
	ChatManager groupChatManager;
	
	Table contactListTable;
	Table eventTable;
	TextArea eventText;
	JabberConnection connection;
	
	Vector contactList;
	Vector eventList;
}