/*
* THIS SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS.  BITFLASH GRAPHICS INC.
* (BITFLASH) ASSUMES NO RESPONSIBILITY FOR THE ACCURACY OR COMPLETENESS OF THE
* DOCUMENTATION AND EXPRESSLY DISCLAIMS ALL REPRESENTATIONS AND WARRANTIES OF ANY
* KIND REGARDING THE CONTENT OR USE OF THE SOFTWARE OR DOCUMENTATION INCLUDING,
* BUT NOT LIMITED TO, EXPRESS AND IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
* FOR A PARTICULAR USE, OR NON-INFRINGEMENT.
*
* IN NO EVENT WILL BITFLASH BE LIABLE FOR ANY DIRECT, INDIRECT, SPECIAL,
* INCIDENTAL OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, LOST BUSINESS OR
* LOST DATA, RESULTING FROM THE USE OF OR RELIANCE UPON THE SOFTWARE OR
* DOCUMENTATION, WHETHER OR NOT BITFLASH HAS BEEN ADVISED OF THE POSSIBILITY OF
* SUCH DAMAGES.
*/

package com.bitflash.jabber.client.gui.whiteboard;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import com.bitflash.jabber.client.gui.*;
import com.bitflash.jabber.client.model.jabber.*;
import com.bitflash.jabber.client.model.svg.*;

public class WhiteboardControl extends Canvas implements MouseListener, MouseMotionListener
{
	static final int TRASH_RADIUS = 15;	
	static final int TRASH_MARGIN = 5;	
	
	// States
	static final int NORMAL = 0;
	static final int MOVE = 1;
	static final int DRAG_HANDLE = 2;
	
	// Transitions
	static final int PEN_DOWN = 0;
	static final int PEN_UP = 1;
	static final int PEN_DRAG = 2;
		
	public WhiteboardControl( ChatWindow parent, ChatManager manager )
	{
		this.parent = parent;
		this.manager = manager;
		this.model = new WhiteboardModel( this, manager );
		this.state = NORMAL;
		
		addMouseListener( this );
		addMouseMotionListener( this );
	}
	
	public void open()
	{
		model.open();
	}
	
	public void close()
	{
		model.close();
	}
	
	public void command( Message message )
	{
		model.command( message );
		repaint();
	}
	
	public void presence( Presence presence )
	{
		model.presence( presence );
		repaint();
	}
	
	public void sendCommand( X whiteboardCommand )
	{
		manager.sendWhiteboardCommand( parent, whiteboardCommand );
	}
		
	public void paint( Graphics context )
	{
		// Highlight the trash can
		if( NORMAL != state && detectTrash(lastPoint) )
		{
			context.setColor( Color.blue );
			context.fillRect(
				0,
				getBounds().height - TRASH_MARGIN - TRASH_RADIUS - TRASH_MARGIN,
				TRASH_MARGIN + TRASH_RADIUS + TRASH_MARGIN,
				TRASH_MARGIN + TRASH_RADIUS + TRASH_MARGIN
			);				
		}
		
		context.setColor( Color.red );
		context.fillOval( 
			TRASH_MARGIN, 
			getBounds().height - TRASH_MARGIN - TRASH_RADIUS, 
			TRASH_RADIUS, 
			TRASH_RADIUS 
		);
		context.setColor( Color.black );
		
		model.paint( context );		
	}
	
	// ------------------------------------------------------- FINITE STATE MACHINE
	
	synchronized private void handle( int transition, Point point )
	{	
		point = boundedPoint( point );
		
		switch( state )
		{
			case NORMAL:
				handleNORMAL( transition, point );
				break;
				
			case MOVE:
				handleMOVE( transition, point );
				break;
				
			case DRAG_HANDLE:
				handleDRAG_HANDLE( transition, point );
				break;
		}
		
		lastPoint = point;
		
		repaint();
	}
	
	private void handleNORMAL( int transition, Point point )
	{
		switch( transition )
		{
			case PEN_DOWN:
			{
				WhiteboardObject object = model.detect( point );
				
				if( null != object )
				{
					object.lock();
					
					if( !object.isSelected() )
					{	
						model.select( object );
						state = MOVE;
						break;
					}
					
					WhiteboardHandle handle = object.detectHandle( point );
					if( null != handle )
					{						
						activeHandle = handle;
						state = DRAG_HANDLE;
						break;
					}
										
					state = MOVE;
					break;
				}
				
				model.unselectAll();
				
				if( detectTrash(point) )
					break;
				
				// TODO: Add more tools here other than lines.
				object = new WhiteboardLine( model, new Line(point, point) );
				object.lock();
				
				model.addObject( object );
				model.select( object );

				// Grab one of the handles and move it around.
				activeHandle = object.detectHandle( point );
				state = DRAG_HANDLE;				
				break;
			}
		}
	}
	
	private void handleMOVE( int transition, Point point )
	{
		switch( transition )
		{
			case PEN_DRAG:
			{				
				int deltaX = point.x - lastPoint.x;
				int deltaY = point.y - lastPoint.y;
				
				Enumeration E = model.getSelected().elements();
				while( E.hasMoreElements() )
				{
					WhiteboardObject object = (WhiteboardObject)E.nextElement();
					object.translate( deltaX, deltaY );
				}
				break;
			}
			
			case PEN_UP:
			{
				state = NORMAL;
				
				if( detectTrash(point) )
				{
					model.deleteSelected();
					break;
				}
				
				model.updateSelected();
				model.unlockSelected();								
				break;
			}
		}
	}
	
	private void handleDRAG_HANDLE( int transition, Point point )
	{
		switch( transition )
		{
			case PEN_DRAG:
			{
				int deltaX = point.x - lastPoint.x;
				int deltaY = point.y - lastPoint.y;
				activeHandle.translate( deltaX, deltaY );
				activeHandle.updateObject();
				break;
			}
			
			case PEN_UP:
			{
				state = NORMAL;
				
				// Check for zero-length lines on creation. This deals
				// with the case where the user is just clicking on the
				// whitespace without trying to draw a line. It's a bit
				// of a hack, but it makes the UI a little easier.
				{
					Vector selected = model.getSelected();
					if( 1 == selected.size() )
					{
						WhiteboardObject object = (WhiteboardObject)selected.firstElement();
						if( !object.isAlive() && object.isSingularity() )
						{
							model.delete( object );
							break;
						}
					}
				}
				
				if( detectTrash(point) )
				{
					model.delete( activeHandle.parent );
					break;
				}
				
				model.updateSelected();
				model.unlockSelected();								
				break;
			}				
		}
	}

	// --------------------------------------------------------------HELPER METHODS
	
	private boolean detectTrash( Point point )
	{
		Rectangle bounds = getBounds();
		return point.x <= TRASH_MARGIN + TRASH_RADIUS + TRASH_MARGIN &&
			bounds.height >= point.y &&
			bounds.height - point.y <= TRASH_MARGIN + TRASH_RADIUS + TRASH_MARGIN;		
	}
	
	private Point boundedPoint( Point point )
	{
		Rectangle bounds = getBounds();
		Point result = new Point(point);
				
		// Keep line inside
		result.x = Math.max( 0, result.x );
		result.x = Math.min( bounds.width-1, result.x );
		result.y = Math.max( 0, result.y );
		result.y = Math.min( bounds.height-1, result.y );
		
		return result;
	}
		
	// --------------------------------------------------------------- MOUSE EVENTS
	
	public void mousePressed( MouseEvent event ) 
	{
		if( (event.BUTTON1_MASK & event.getID()) != 0 )
			handle( PEN_DOWN, event.getPoint() );
	}

	public void mouseReleased( MouseEvent event ) 
	{
		if( (event.BUTTON1_MASK & event.getID()) != 0 )
			handle( PEN_UP, event.getPoint() );
	}	
	
	public void mouseDragged( MouseEvent event )
	{
		if( (event.BUTTON1_MASK & event.getID()) != 0 )
			handle( PEN_DRAG, event.getPoint() );		
	}
			
	public void mouseClicked( MouseEvent event ) {}	
	public void mouseEntered( MouseEvent event ) {}
	public void mouseExited( MouseEvent event ) {}	
	public void mouseMoved( MouseEvent event ) {}	
	
	ChatWindow parent;
	ChatManager manager;
	
	WhiteboardModel model;
	
	Point lastPoint;
	WhiteboardHandle activeHandle;
	
	int state;
}
