/*
* THIS SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS.  BITFLASH GRAPHICS INC.
* (BITFLASH) ASSUMES NO RESPONSIBILITY FOR THE ACCURACY OR COMPLETENESS OF THE
* DOCUMENTATION AND EXPRESSLY DISCLAIMS ALL REPRESENTATIONS AND WARRANTIES OF ANY
* KIND REGARDING THE CONTENT OR USE OF THE SOFTWARE OR DOCUMENTATION INCLUDING,
* BUT NOT LIMITED TO, EXPRESS AND IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
* FOR A PARTICULAR USE, OR NON-INFRINGEMENT.
*
* IN NO EVENT WILL BITFLASH BE LIABLE FOR ANY DIRECT, INDIRECT, SPECIAL,
* INCIDENTAL OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, LOST BUSINESS OR
* LOST DATA, RESULTING FROM THE USE OF OR RELIANCE UPON THE SOFTWARE OR
* DOCUMENTATION, WHETHER OR NOT BITFLASH HAS BEEN ADVISED OF THE POSSIBILITY OF
* SUCH DAMAGES.
*/

package com.bitflash.jabber.client.gui.whiteboard;

import java.awt.*;
import java.util.*;
import com.bitflash.jabber.client.model.*;
import com.bitflash.jabber.client.model.jabber.*;
import com.bitflash.jabber.client.model.svg.*;
import com.bitflash.jabber.client.gui.ChatManager;

/**
 * Controls the data model of the Whiteboard. The whiteboard
 * data model has more to it than the underlying SVG tree;
 * it includes the state of the whiteboard object, plus the
 * synchronization with the foreign clients. This class 
 * essentially implements the whiteboard protocol.
 */
public class WhiteboardModel
{
	static final String SELECT = "select";
	static final String UNSELECT = "unselect";
	static final String UPDATE = "update";
	static final String DELETE = "delete";
	static final String SYNCHRONIZE = "synchronize";
	static final String ACKNOWLEDGE = "acknowledge";
	static final String SUBSCRIBE = "subscribe";
	static final String UNSUBSCRIBE = "unsubscribe";
	
	/**
	 * @param jabberID The Jabber ID of the local user.
	 */
	public WhiteboardModel( WhiteboardControl control, ChatManager manager )
	{
		this.control = control;
		this.manager = manager;
		this.liveObjects = new Vector();
		this.foreignClients = new Vector();
	}
	
	public void open()
	{
		send( SUBSCRIBE );
		send( SYNCHRONIZE );					
	}
	
	public void close()
	{
		send( UNSUBSCRIBE );
	}
	
	public void command( Message message )
	{
		String from = (String)message.getProperty("from");
		
		// Don't listen to commands from ourselves.
		if( manager.nickname(from).equals( manager.bareNickname(manager.jabberID()) ) )
			return;
		
		// Don't listen to delayed messages (e.g. those cached in the chat history)
		if( null != message.getChild( "x", X.XMLNS_DELAY ) )
			return;
		
		X x = (X)message.getChild( "x", X.XMLNS_WHITEBOARD );
		SVG svg = (SVG)x.getProperty("svg");
		
		// Ignore messages addressed to others.
		String to = manager.nickname((String)x.getProperty("to"));
		if( !to.equals("") && !to.equals(manager.bareNickname(manager.jabberID())) )
			return;
		
		String action = ((String)x.getProperty( "action" )).toLowerCase();
									
		// If svg is null, extractRecognizedSVG magically returns a Vector
		// with a single null in it. It makes the code easier, though.
		Enumeration E = extractRecognizedSVG( svg ).elements();
		while( E.hasMoreElements() )
			commandSingle( action, (DocumentObject)E.nextElement(), from );
				
		// Acknowledge the whole UPDATE action at once.
		if( action.equals(UPDATE) )
			ACKNOWLEDGE( to, svg );
	}
	
	private void commandSingle( String action, DocumentObject xmlElement, String from )
	{		
		if( action.equals(SUBSCRIBE) )
			commandSUBSCRIBE( from );

		if( action.equals(UNSUBSCRIBE) )
			commandUNSUBSCRIBE( from );

		if( action.equals(SELECT) && null != xmlElement )
			commandSELECT( xmlElement, from );
		
		else if( action.equals(UNSELECT) && null != xmlElement )
			commandUNSELECT( xmlElement, from );

		else if( action.equals(UPDATE) && null != xmlElement )
			commandUPDATE( xmlElement, from );
		
		else if( action.equals(DELETE) && null != xmlElement )
			commandDELETE( xmlElement, from );
		
		else if( action.equals(ACKNOWLEDGE) && !from.equals("") )
			commandACKNOWLEDGE( xmlElement, from );
		
		else if( action.equals(SYNCHRONIZE) && !from.equals("") )
			commandSYNCHRONIZE( from );
	}	
	
	public void presence( Presence presence )
	{
		String type = ((String)presence.getProperty("type")).toLowerCase();
		String from = (String)presence.getProperty("from");			
		
		// Who?!
		if( from.equals("") )
			return;
				
		 if( type.equals("unavailable") )
			commandUNSUBSCRIBE( from );
	}
		
	/**
	 * Adds the object to the whiteboard. The object will be put
	 * into the newborn state. It must be update()d before it
	 * becomes alive. 
	 */
	public void addObject( WhiteboardObject object )
	{					
		object.svgObject.setProperty( "id", generateID() );
		object.synchronizationState = WhiteboardObject.stateNEWBORN;
		liveObjects.addElement( object );
	}
	
	public void select( WhiteboardObject object )
	{
		// Don't do unnecessary work
		if( !object.isSelected() )
		{
			object.isSelected = true;
			send( SELECT, object );
		}
	}
	
	public void unselectAll()
	{	
		Vector selected = getSelected();
		Enumeration E = selected.elements();
		while( E.hasMoreElements() )
		{
			WhiteboardObject object = (WhiteboardObject)E.nextElement();
			object.isSelected = false;
		}
		
		send( UNSELECT, selected );
	}		
	
	public Vector getSelected()
	{
		Vector result = new Vector();
		Enumeration E = liveObjects.elements();
		while( E.hasMoreElements() )
		{
			WhiteboardObject object = (WhiteboardObject)E.nextElement();
			if( object.isSelected() )
				result.addElement( object );
		}		
		return result;
	}
	
	/**
	 * This automagically unlocks and unselects the objects too.
	 */
	public void deleteSelected()
	{
		Vector selected = getSelected();
		Enumeration E = selected.elements();
		while( E.hasMoreElements() )
		{
			WhiteboardObject object = (WhiteboardObject)E.nextElement();			
			doDelete( object );
		}
		
		send( DELETE, selected );
	}

	/**
	 * This automagically unlocks and unselects the object too.
	 */	
	public void delete( WhiteboardObject object )
	{
		if( !liveObjects.contains(object) )
			return;
		
		doDelete( object );
		
		send( DELETE, object );
	}		
	
	public void updateSelected()
	{
		Vector selected = getSelected();
		Vector newSelect = new Vector();
		
		// Because we are sending the object across the wire, we 
		// now know that the foreign clients are aware that it
		// exists. So, the object is alive.
		Enumeration E = selected.elements();
		while( E.hasMoreElements() )
		{
			WhiteboardObject object = (WhiteboardObject)E.nextElement();			
			object.addForeignAcknowledgments( foreignClients );
			
			if( !object.isAlive() )
			{
				object.synchronizationState = WhiteboardObject.stateALIVE;
				
				// Make sure we send the select status if we are sending
				// this over the wire for the first time.
				if( object.isSelected() )
					newSelect.addElement( object );
			}
		}
		
		send( "", UPDATE, selected, true );
		send( SELECT, newSelect );
	}
	
	public void update( WhiteboardObject object )
	{
		int previousState = object.synchronizationState;
		object.synchronizationState = WhiteboardObject.stateALIVE;
		
		object.addForeignAcknowledgments( foreignClients );
		send( "", UPDATE, object, true );
		
		// Make sure we send the select status if we are sending
		// this over the wire for the first time.
		if( previousState != object.synchronizationState && object.isSelected() )
			send( SELECT, object );
	}
	
	public void unlockSelected()
	{
		Enumeration E = liveObjects.elements();
		while( E.hasMoreElements() )
		{
			WhiteboardObject object = (WhiteboardObject)E.nextElement();
			if( object.isSelected() )
				object.unlock();
		}				
	}
	
	/**
	 * Detects the object at the given point. Returns null 
	 * if there is no such object.
	 */
	public WhiteboardObject detect( Point point )
	{
		Enumeration E = liveObjects.elements();
		while( E.hasMoreElements() )
		{
			WhiteboardObject object = (WhiteboardObject)E.nextElement();
			if( object.detect( point ) )
				return object;
		}		

		return null;		
	}
	
	public void paint( Graphics context )
	{
		Enumeration E = liveObjects.elements();
		while( E.hasMoreElements() )
		{
			WhiteboardObject object = (WhiteboardObject)E.nextElement();		
			object.paint( context );
		}
	}
		
	private String generateID()
	{
		return manager.bareNickname(manager.jabberID()) + '/' + Integer.toString( ++lastID );
	}
	
	private void send( String action )
	{
		send( action, (WhiteboardObject)null );
	}
	
	private void send( String action, WhiteboardObject object )
	{
		send( "", action, object, false );
	}
	
	private void send( String to, String action, WhiteboardObject object, boolean sendComplete )
	{		
		SVG svg = null;
		if( null != object ) 
		{
			// Nothing to send.
			if( WhiteboardObject.stateNEWBORN == object.synchronizationState )
				return;
			
			svg = new SVG();
			svg.addChild( sendComplete ? object.svgObject : object.svgObject.idOnly() );
		}
		
		send( to, action, svg );
	}

	private void send( String action, Vector objects )
	{
		send( "", action, objects, false );
	}	
	
	private void send( String to, String action, Vector objects, boolean sendComplete )
	{
		SVG svg = new SVG();
		
		Enumeration E = objects.elements();
		while( E.hasMoreElements() )
		{
			WhiteboardObject object = (WhiteboardObject)E.nextElement();
			if( WhiteboardObject.stateNEWBORN != object.synchronizationState )			
				svg.addChild( sendComplete ? object.svgObject : object.svgObject.idOnly() );
		}

		// Nothing to send.
		if( svg.children.isEmpty() )
			return;
		
		send( to, action, svg );		
	}
		
	private void send( String to, String action, SVG svg )
	{
		X x = new X( X.XMLNS_WHITEBOARD );
		x.setProperty( "action", action );
		x.setProperty( "to", to );	
		
		if( null != svg )
			x.setProperty( "svg", svg );
		
		control.sendCommand( x );
	}
	
	private void ACKNOWLEDGE( String to, SVG svg )
	{
		SVG reduced = new SVG();
		reduced.setProperties(svg);
		
		Enumeration E = svg.children.elements();
		while( E.hasMoreElements() )
			reduced.addChild( ((DocumentObject)E.nextElement()).idOnly() );
		
		send( to, ACKNOWLEDGE, reduced );
	}
	
	// ----------------------------------------------------- COMMAND HANDLERS

	private void commandSUBSCRIBE( String from )
	{
		// If we're in private chat, it's possible the client didn't
		// say goodbye or change presence to unavailable. Thus, we shall 
		// clean up.
		commandUNSUBSCRIBE( from );
		
		if( !foreignClients.contains(from) )		
			foreignClients.addElement(from);
		
		send( from, ACKNOWLEDGE, null );
	}
	
	private void commandUNSUBSCRIBE( String from )
	{
		if( foreignClients.contains(from) )
			foreignClients.removeElement( from );
		
		Enumeration live = liveObjects.elements();
		while( live.hasMoreElements() )
		{
			WhiteboardObject object = (WhiteboardObject)live.nextElement();
			object.foreignAcknowledgeAll( from );
			object.foreignUnselect( from );
		}
	}
				
	private void commandSELECT( DocumentObject xmlElement, String from )
	{		
		WhiteboardObject object = findObject( (String)xmlElement.getProperty("id") );
		
		// Unknown object.
		if( null == object )
			return;
		
		object.foreignSelect( from );
	}
	
	private void commandUNSELECT( DocumentObject xmlElement, String from )
	{
		WhiteboardObject object = findObject( (String)xmlElement.getProperty("id") );
		
		// Unknown object.
		if( null == object )
			return;
		
		object.foreignUnselect( from );		
	}	
	
	private void commandUPDATE( DocumentObject xmlElement, String from )
	{
		WhiteboardObject object = findObject( (String)xmlElement.getProperty("id") );
			
		// Unknown object.
		if( null == object )
		{
			object = compile( xmlElement );
			liveObjects.addElement( object );

			// Don't duplicate IDs. That would be bad.
			try
			{
				String me = manager.bareNickname(manager.jabberID());			
				String id = (String)xmlElement.getProperty("id");
				int slash = id.lastIndexOf('/');
				if( slash > 0 && id.substring(0, slash).equals(me) )
					lastID = Math.max( lastID, Integer.parseInt(id.substring(slash+1)) );
			}
			catch( NumberFormatException e )
			{
				// ID not in our namespace, so we don't care.
			}
			catch( IndexOutOfBoundsException e )
			{
				// ID not in our namespace, so we don't care.
			}			
		}
		else if( object.isForeignDirty() || object.isLocked() )
			fork( object );
			
		else
			object.svgObject.setProperties( xmlElement );
	}
	
	private void commandDELETE( DocumentObject xmlElement, String from )
	{
		WhiteboardObject object = findObject( (String)xmlElement.getProperty("id") );
			
		// Unknown object.
		if( null == object )
			return;

		else if( object.isForeignDirty() || object.isLocked() || object.isSelected() )
			fork( object );
			
		else
			liveObjects.removeElement( object );			
	}	
	
	private void commandACKNOWLEDGE( DocumentObject xmlElement, String from )
	{
		// Acknowledge from SUBSCRIBE
		if( null == xmlElement )
		{
			if( !foreignClients.contains(from) )		
				foreignClients.addElement(from);
			return;			
		}
		
		// Else Acknowledge from UPDATE		
		String id = (String)xmlElement.getProperty("id");		
		WhiteboardObject object = findObject( id );
		
		// Unknown object.
		if( null == object )
			return;
				
		object.foreignAcknowledge( from );
	}
	
	private void commandSYNCHRONIZE( String from )
	{
		send( from, UPDATE, liveObjects, true );
		send( from, SELECT, getSelected(), false );
	}
	
	// ----------------------------------------------------------------------
	
	/**
	 * This deletes the object in question and creates a new one
	 * with a new (local) ID and the same configuration. If the
	 * object is selected, the forked object is selected as well.
	 * 
	 * In reality, a fork operation is two or three commands:
	 * 
	 * DELETE object
	 * UPDATE fork
	 * [SELECT fork] if object was selected
	 */
	private WhiteboardObject fork( WhiteboardObject object )
	{					
		WhiteboardObject forked = object.fork( generateID() );
		liveObjects.addElement( forked );
		
		boolean isSelected = object.isSelected();
		delete( object );
				
		update( forked );
		if( isSelected )
			select( forked );
		
		return forked;
	}
	
	/**
	 * Returns a collection of all objects inside the SVG document
	 * that we recognize on this whiteboard.<P>
	 * 
	 * If svg is null, return a collection with a single null in it.
	 * While this is bizarre, it makes the code above easier. 
	 * 
	 * For now, we are <em>very</em> cheap, and only recognize lines 
	 * that are direct children of the root <svg/> tag. For demonstration
	 * purposes this is acceptable. But for production, really, we'd need
	 * to go through the whole SVG DOM tree looking for objects that are
	 * embedded inside of groups, etc. that we recognize.
	 */
	private Vector extractRecognizedSVG( SVG svg )
	{
		if( null == svg )
		{
			Vector result = new Vector();
			result.addElement(null);
			return result;
		}
		return svg.getChildren( "line" );
	}
	
	/**
	 * Finds the corresponding WhiteboardObject by ID. It will return
	 * null if the ID is "" (i.e. there is no ID), or no object with
	 * the given ID is found. Objects on the deleted list won't be searched.
	 * See findDeletedObject() for that.
	 */
	private WhiteboardObject findObject( String id )
	{
		Enumeration E = liveObjects.elements();
		while( E.hasMoreElements() )
		{
			WhiteboardObject object = (WhiteboardObject)E.nextElement();
			if( object.svgObject.getProperty("id").equals(id) )
				return object;
		}
		return null;
	}
	
	/**
	 * Returns a new WhiteboardObject based on the given SVG element.
	 * null on failure.
	 */
	private WhiteboardObject compile( DocumentObject object )
	{		
		// So far, we only recognize lines. That's because we are cheap.
		if( Line.class == object.getClass() )
			return new WhiteboardLine(this, (Line)object);
		
		return null;
	}
	
	/**
	 * The core of the delete functionality, extracted out because it
	 * is non-trivial. This automagically clears the selected and lock
	 * flags.
	 */
	private void doDelete( WhiteboardObject object )
	{
		liveObjects.removeElement( object );
					
		// Unselect quietly
		object.isSelected = false;
		object.unlock();
		object.synchronizationState = object.stateDELETED;
	}
						  		
	WhiteboardControl control;	
	ChatManager manager;
	
	Vector liveObjects;
	Vector foreignClients;
		
	int lastID = 0;
}