/*
* THIS SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS.  BITFLASH GRAPHICS INC.
* (BITFLASH) ASSUMES NO RESPONSIBILITY FOR THE ACCURACY OR COMPLETENESS OF THE
* DOCUMENTATION AND EXPRESSLY DISCLAIMS ALL REPRESENTATIONS AND WARRANTIES OF ANY
* KIND REGARDING THE CONTENT OR USE OF THE SOFTWARE OR DOCUMENTATION INCLUDING,
* BUT NOT LIMITED TO, EXPRESS AND IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
* FOR A PARTICULAR USE, OR NON-INFRINGEMENT.
*
* IN NO EVENT WILL BITFLASH BE LIABLE FOR ANY DIRECT, INDIRECT, SPECIAL,
* INCIDENTAL OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, LOST BUSINESS OR
* LOST DATA, RESULTING FROM THE USE OF OR RELIANCE UPON THE SOFTWARE OR
* DOCUMENTATION, WHETHER OR NOT BITFLASH HAS BEEN ADVISED OF THE POSSIBILITY OF
* SUCH DAMAGES.
*/

package com.bitflash.jabber.client.gui.whiteboard;

import java.awt.*;
import java.util.*;
import java.lang.reflect.*;
import com.bitflash.jabber.client.model.*;

/** 
 * Base class of all the things that may exist on the whiteboard.
 * Child constructors must match the signature 
 * 
 *    ChildObject( WhiteboardModel, DocumentObject )
 * 
 * or fork() will fail.
 */
public abstract class WhiteboardObject
{	
	public WhiteboardObject( WhiteboardModel model, DocumentObject svgObject )
	{
		this.model = model;	
		this.svgObject = svgObject;
		
		this.foreignSelected = new Vector();
		this.foreignAcknowledgments = new Vector();
		this.handles = new Vector();
	}
	
	/** 
	 * If any foreign client changes this object, the model
	 * will fork this object. This prevents any changes to
	 * be made. For instance, when dragging a line, if someone
	 * grabs it and moves it on you, the object will fork.
	 */
	public void lock()
	{
		this.isLocked = true;
	}
	
	public void unlock()
	{
		this.isLocked = false;
	}
	
	public boolean isLocked()
	{
		return isLocked;
	}
		
	/**
	 * Indicates whether the current object is selected
	 * by the local client. Selection prevents deletion,
	 * and it also indicates that the local author is 
	 * doing work on the object. For stricter lock outs--
	 * such that no changes can be made--see setLocked()/isLocked()<P>
	 * 
	 * To change the selected status of this object, go through
	 * WhiteboardModel.select().
	 */
	
	public boolean isSelected()
	{
		return isSelected;
	}
	
	/**
	 * An object is alive if it has been sent over the wire to the
	 * other whiteboard peers, and if it hasn't been deleted. An
	 * object can also be newborn (as in created, but not sent
	 * over the wire), or dead (deleted).
	 */
	public boolean isAlive()
	{
		return stateALIVE == synchronizationState;
	}	

	/**
	 * Keeps track of which foreign clients have selected 
	 * this object.
	 */
	public void foreignSelect( String jabberID )
	{
		if( !foreignSelected.contains(jabberID) )
			foreignSelected.addElement( jabberID );
	}
	
	public void foreignUnselect( String jabberID )
	{
		if( foreignSelected.contains(jabberID) )
			foreignSelected.removeElement( jabberID );
	}
	
	public boolean isForeignSelected()
	{
		return !foreignSelected.isEmpty();
	}
	
	/**
	 * Keeps track of which foreign clients need to acknowledge
	 * any pending changes.
	 */
	public void addForeignAcknowledgments( Vector jabberIDs )
	{
		Enumeration E = jabberIDs.elements();
		while( E.hasMoreElements() )
			foreignAcknowledgments.addElement( E.nextElement() );
	}
	
	public void foreignAcknowledge( String jabberID )
	{
		if( foreignAcknowledgments.contains(jabberID) )
			foreignAcknowledgments.removeElement( jabberID );
	}
	
	public void foreignAcknowledgeAll( String jabberID )
	{
		while( foreignAcknowledgments.contains(jabberID) )
			foreignAcknowledgments.removeElement( jabberID );
	}	
	
	public boolean isForeignDirty()
	{
		return !foreignAcknowledgments.isEmpty();
	}
	
	/**
	 * Creates a similar (but not clone()d) object, with the given ID.
	 * @return a new WhiteboardObject, or null on error.
	 */
	public WhiteboardObject fork( String id )
	{
		WhiteboardObject result;
		try {
			DocumentObject resultSVG = fork( svgObject );		
			resultSVG.setProperty( "id", id );				
			
			Class objectClass = getClass();
			Constructor constructor = objectClass.getConstructor( new Class[] { 
				WhiteboardModel.class, 
				DocumentObject.class 
			} );
			result = (WhiteboardObject)constructor.newInstance( new Object[] { model, resultSVG } ); 
			
		} catch( Throwable t ) {
			t.printStackTrace();
			return null;
		}		
						
		return result;
	}

	/**
	 * Forks a DocumentObject and its children.
	 */
	private DocumentObject fork( DocumentObject object )
	{
		DocumentObject result;
		try {
			result = (DocumentObject)object.getClass().newInstance();
		} catch( Throwable t ) {
			t.printStackTrace();
			return null;
		}		
		
		Enumeration E = object.children.elements();
		while( E.hasMoreElements() )
		{
			DocumentObject child = (DocumentObject)E.nextElement();
			result.addChild( fork(child) );
		}
		
		result.setProperties( object );
		return result;
	}
	
	/**
	 * Returns true if the point is on this object on the whiteboard.
	 */
	abstract public boolean detect( Point point );
	
	/**
	 * Returns the appropriate WhiteboardHandle given a point.
	 * Returns null if the point is not on a handle.
	 */
	public WhiteboardHandle detectHandle( Point point )
	{
		updateHandles();
		
		Enumeration E = handles.elements();
		while( E.hasMoreElements() )
		{
			WhiteboardHandle handle = (WhiteboardHandle)E.nextElement();
			if( handle.detect(point) )
				return handle;
		}
		return null;
	}
	
	/**
	 * Returns true if the object's extent in all directions is zero
	 * (or one pixel).
	 */
	abstract public boolean isSingularity();
	
	/**
	 * For all vertices (or significant points), translate them by the 
	 * given deltas. So, for a vertex v = (x,y), 
	 * 
	 * <BLOCKQUOTE>
	 *       v => v' = (x + deltaX, y + deltaY); 
	 * </BLOCKQUOTE><P>
	 * 
	 * This uses the handles to do the work. In other words, all <STRONG>handles</STRONG>
	 * will be translated.
	 */
	public void translate( int deltaX, int deltaY )
	{
		updateHandles();
		
		Enumeration E = handles.elements();
		while( E.hasMoreElements() )
		{
			WhiteboardHandle handle = (WhiteboardHandle)E.nextElement();
			handle.translate( deltaX, deltaY );
			handle.updateObject();
		}
	}
	
	/**
	 * Handle management. In order for the handles to work, subclasses
	 * must register them when they are instantiated and then implement 
	 * updateHandles() and updateFromHandle().
	 */
	protected void registerHandle( WhiteboardHandle handle )
	{
		handles.addElement( handle );
		handle.parent = this;
	}
	
	/**
	 * Puts the correct information in the handle.
	 */
	abstract protected void updateHandles();
	abstract void updateFromHandle( WhiteboardHandle handle );
	
	/**
	 * Draws the object.
	 */
	abstract public void paint( Graphics context );
	
	/**
	 * The SVG object that this thing represents.
	 */
	DocumentObject svgObject;
	
	/**
	 * The WhiteboardModel controlling this object.
	 */
	WhiteboardModel model;
	
	boolean isLocked;
	
	// This is updated by the WhiteboardModel, primarily (exception: fork())
	boolean isSelected; 
	
	Vector foreignSelected;
	Vector foreignAcknowledgments;
	Vector handles;
	
	/**
	 * The state according to the WhiteboardModel. This is maintained
	 * exclusively by and for the model to handle the distributed
	 * whiteboard protocol. (mostly as an optimization)
	 */
	
	static final int stateNEWBORN = 0;
	static final int stateALIVE = 1;
	static final int stateDELETED = 2;
	int synchronizationState = stateALIVE;
}
