/*
* THIS SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS.  BITFLASH GRAPHICS INC.
* (BITFLASH) ASSUMES NO RESPONSIBILITY FOR THE ACCURACY OR COMPLETENESS OF THE
* DOCUMENTATION AND EXPRESSLY DISCLAIMS ALL REPRESENTATIONS AND WARRANTIES OF ANY
* KIND REGARDING THE CONTENT OR USE OF THE SOFTWARE OR DOCUMENTATION INCLUDING,
* BUT NOT LIMITED TO, EXPRESS AND IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
* FOR A PARTICULAR USE, OR NON-INFRINGEMENT.
*
* IN NO EVENT WILL BITFLASH BE LIABLE FOR ANY DIRECT, INDIRECT, SPECIAL,
* INCIDENTAL OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, LOST BUSINESS OR
* LOST DATA, RESULTING FROM THE USE OF OR RELIANCE UPON THE SOFTWARE OR
* DOCUMENTATION, WHETHER OR NOT BITFLASH HAS BEEN ADVISED OF THE POSSIBILITY OF
* SUCH DAMAGES.
*/

package com.bitflash.jabber.client.xml;

import java.io.*;
import org.xml.sax.*;
import org.xml.sax.helpers.AttributeListImpl;
import java.util.StringTokenizer;

/**
 * A tiny, almost-SAX parser for streaming XML.
 */
public class StreamingXMLParser
{	
	static final boolean TRACE = false;
	static int traceLength = 0;
	
	static final String CHARACTER_ENTITY_TRANSLATIONS[] = {
		"amp",				"&",
		"lt",				"<",
		"gt",				">",
		"quot",				"\"",
		"apos",				"'",
	};
	
	public StreamingXMLParser( InputStream input, DocumentHandler handler )
	{		
		this.input = new PushbackInputStream(input, 512);
		this.handler = handler;		
		state = PICKUP;
		quoteCharacter = '\0';
		emptyBuffer();
	}
	
	public void start() throws SAXException, IOException
	{
		handler.startDocument();

		try
		{		
			int c;
			while( (c = input.read()) > 0 )
				parse((char)c);
			
			if( buffer.length() > 0 )
				sendCharacters();			
		}
		catch( Throwable t )
		{
			t.printStackTrace();
		}
			
		handler.endDocument();
	}
	
	static public String encodeUTF( String source )
	{
		StringBuffer result = new StringBuffer();
		byte bytes[] = source.getBytes();
		
		for( int i = 0; i < bytes.length; i++ )
		{		
			int j;
			for( j = 1; j < CHARACTER_ENTITY_TRANSLATIONS.length; j += 2 )
			{
				if( bytes[i] == CHARACTER_ENTITY_TRANSLATIONS[j].charAt(0) )
				{
					result.append( '&' );
					result.append( CHARACTER_ENTITY_TRANSLATIONS[j-1] );
					result.append( ';' );
					break;
				}
			}
			
			if( j >= CHARACTER_ENTITY_TRANSLATIONS.length )
				result.append( (char)bytes[i] );
		}
		return result.toString();
	}
	
	static public String decodeUTF( String source ) throws SAXException
	{
		StringBuffer result = new StringBuffer();
		
		int current = 0;
		
		while( current < source.length() )
		{
			int amp = source.indexOf( "&", current );
			if( amp < 0 )
			{
				result.append( source.substring(current) );
				break;
			}
			else
			{
				result.append( source.substring(current, amp) );
				
				int semicolon = source.indexOf( ";", current );
				if( semicolon < 0 )
					throw new SAXException( "Malformed escape at fragment: " + source.substring(current) );
				
				String escape = source.substring( amp + 1, semicolon ).toLowerCase();
				current = semicolon + 1;
				
				if( '#' == escape.charAt(0) )
				{
					if( 'x' == escape.charAt(1) )
						result.append( (char)Integer.parseInt(escape.substring(2), 16) );						
					
					else
						result.append( (char)Integer.parseInt(escape.substring(1), 10) );
				}
				else
				{
					String translation = "?";
					for( int i = 0; i < CHARACTER_ENTITY_TRANSLATIONS.length; i += 2 )
						if( escape.equals(CHARACTER_ENTITY_TRANSLATIONS[i]) )
						{
							translation = CHARACTER_ENTITY_TRANSLATIONS[i+1];
							break;
						}
					
					result.append( translation );
				}					
			}			
		}	
		return result.toString();
	}	
	
	protected void parse( char c ) throws SAXException
	{
		if( TRACE )
		{
			System.out.write(c);
			if( ++traceLength == 80 )
			{
				traceLength = 0;
				System.out.write('\n');
			}
			System.out.flush();
		}
		
		switch( state )
		{
			case PICKUP:
			{
				if( '<' == c ) {
					state = TAG;
					sendCharacters();
					emptyBuffer();
				} else {
					buffer.append(c);
				}
				break;
			}
			
			// Just read everything inbetween the two brackets, 
			// mindful of quotation marks.
			case TAG:
			{
				if( '\0' == quoteCharacter && '>' == c )
					parseTag();
					
				else
					buffer.append(c);
				
				if( '"' == c || '\'' == c )
				{
					if( quoteCharacter == c )
						quoteCharacter = '\0';
					else if( quoteCharacter == '\0' )
						quoteCharacter = c;
				}					
			}
		}
	}
	
	protected void parseTag() throws SAXException
	{
		String tag = buffer.toString();
				
		if( '?' == tag.charAt(0) )
		{						
			StringTokenizer tokenizer = new StringTokenizer(tag);
			String name = tokenizer.nextToken();
			String rest = tag.substring(name.length()).trim();
			handler.processingInstruction( name.substring(1), rest );		
		}
		
		else
		{			
			boolean startTag = true;
			boolean endTag = false;
			if( tag.endsWith("/") )
			{
				endTag = true;
				tag = tag.substring(0,tag.length()-1);
			}
			if( tag.startsWith("/") )
			{
				startTag = false;
				endTag = true;
				tag = tag.substring(1);
			}
			tag.trim();
			
			StringTokenizer tokenizer = new StringTokenizer(tag);
			String name = tokenizer.nextToken();
			
			AttributeListImpl attributeList = new AttributeListImpl();
			while( tokenizer.hasMoreTokens() )
			{
				String token = tokenizer.nextToken();
				String attribute = token;
				String value = "";
				
				int equals = token.indexOf('=');
				if( equals > 0 )
				{
					attribute = token.substring(0, equals);
					value = token.substring(equals+1);
					if( '"' == value.charAt(0) || '\'' == value.charAt(0) )
						value = value.substring(1,value.length()-1);
				}
				
				attributeList.addAttribute( attribute, "CDATA", value );
			}
			
			if( startTag )
				handler.startElement( name, attributeList );
			if( endTag )
				handler.endElement(name);
		}
		
		state = PICKUP;
		emptyBuffer();		
	}

	protected void emptyBuffer() 
	{ 
		buffer = new StringBuffer( 1024 ); 
	}
	
	protected void sendCharacters() throws SAXException
	{
		if( buffer.length() > 0 )
		{
			char buf[] = new char[buffer.length()];
			buffer.toString().getChars( 0, buffer.length(), buf, 0 );
			handler.characters( buf, 0, buffer.length() );
		}
	}
	
	protected InputStream input;
	protected DocumentHandler handler;
	protected StringBuffer buffer;
	
	private static final int PICKUP = 0;
	private static final int TAG = 1;
	private int state;
	
	private char quoteCharacter;
}
